# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Shared resource flags for kms resources."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope.concepts import concepts
from googlecloudsdk.calliope.concepts import deps
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs
from googlecloudsdk.core import properties


def KeyVersionAttributeConfig(kms_prefix=True):
  name = 'kms-key-version' if kms_prefix else 'key-version'
  return concepts.ResourceParameterAttributeConfig(
      name=name, help_text='The KMS key version of the {resource}.')


def KeyAttributeConfig(kms_prefix=True):
  name = 'kms-key' if kms_prefix else 'key'
  return concepts.ResourceParameterAttributeConfig(
      name=name, help_text='The KMS key of the {resource}.')


def KeyringAttributeConfig(kms_prefix=True):
  name = 'kms-keyring' if kms_prefix else 'keyring'
  return concepts.ResourceParameterAttributeConfig(
      name=name, help_text='The KMS keyring of the {resource}.')


def EkmConnectionAttributeConfig(kms_prefix=True):
  name = 'kms-ekmconnection' if kms_prefix else 'ekmconnection'
  return concepts.ResourceParameterAttributeConfig(
      name=name, help_text='The KMS ekm connection of the {resource}.')


def LocationAttributeConfig(kms_prefix=True, region_fallthrough=False):
  name = 'kms-location' if kms_prefix else 'location'
  fallthroughs = []
  if region_fallthrough:
    fallthroughs.append(deps.ArgFallthrough('--region'))
  return concepts.ResourceParameterAttributeConfig(
      name=name,
      help_text='The Cloud location for the {resource}.',
      fallthroughs=fallthroughs)


def ProjectAttributeConfig(kms_prefix=True):
  name = 'kms-project' if kms_prefix else 'project'
  return concepts.ResourceParameterAttributeConfig(
      name=name,
      help_text='The Cloud project for the {resource}.',
      fallthroughs=[deps.PropertyFallthrough(properties.VALUES.core.project)])


def GetKmsKeyVersionResourceSpec(kms_prefix=True):
  return concepts.ResourceSpec(
      'cloudkms.projects.locations.keyRings.cryptoKeys.cryptoKeyVersions',
      resource_name='key version',
      cryptoKeyVersionsId=KeyVersionAttributeConfig(kms_prefix),
      cryptoKeysId=KeyAttributeConfig(kms_prefix),
      keyRingsId=KeyringAttributeConfig(kms_prefix),
      locationsId=LocationAttributeConfig(kms_prefix=kms_prefix),
      projectsId=ProjectAttributeConfig(kms_prefix=kms_prefix),
      disable_auto_completers=False)


def GetKmsKeyResourceSpec(kms_prefix=True, region_fallthrough=False):
  return concepts.ResourceSpec(
      'cloudkms.projects.locations.keyRings.cryptoKeys',
      resource_name='key',
      cryptoKeysId=KeyAttributeConfig(kms_prefix),
      keyRingsId=KeyringAttributeConfig(kms_prefix),
      locationsId=LocationAttributeConfig(
          kms_prefix=kms_prefix, region_fallthrough=region_fallthrough),
      projectsId=ProjectAttributeConfig(kms_prefix),
      disable_auto_completers=False)


def GetKmsKeyRingResourceSpec(kms_prefix=True, region_fallthrough=False):
  return concepts.ResourceSpec(
      'cloudkms.projects.locations.keyRings',
      resource_name='keyring',
      keyRingsId=KeyringAttributeConfig(kms_prefix),
      locationsId=LocationAttributeConfig(
          kms_prefix=kms_prefix, region_fallthrough=region_fallthrough),
      projectsId=ProjectAttributeConfig(kms_prefix),
      disable_auto_completers=False)


def GetKmsEkmConnectionResourceSpec(kms_prefix=True, region_fallthrough=False):
  return concepts.ResourceSpec(
      'cloudkms.projects.locations.ekmConnections',
      resource_name='ekmconnection',
      ekmConnectionsId=EkmConnectionAttributeConfig(kms_prefix),
      locationsId=LocationAttributeConfig(
          kms_prefix=kms_prefix, region_fallthrough=region_fallthrough),
      projectsId=ProjectAttributeConfig(kms_prefix),
      disable_auto_completers=False)


def GetKmsLocationResourceSpec(kms_prefix=True, region_fallthrough=False):
  return concepts.ResourceSpec(
      'cloudkms.projects.locations',
      resource_name='location',
      locationsId=LocationAttributeConfig(
          kms_prefix=kms_prefix, region_fallthrough=region_fallthrough),
      projectsId=ProjectAttributeConfig(kms_prefix),
      disable_auto_completers=False)


def GetKmsKeyPresentationSpec(resource,
                              region_fallthrough=False,
                              flag_overrides=None,
                              permission_info=None):
  """Return a Presentation Spec for kms key resource argument.

  Args:
    resource: str, the name of the resource that the cryptokey will be used to
    protect.
    region_fallthrough: bool, True if the command has a region flag that should
      be used as a fallthrough for the kms location.
    flag_overrides: dict, The default flag names are 'kms-key', 'kms-keyring',
      'kms-location' and 'kms-project'. You can pass a dict of overrides where
      the keys of the dict are the default flag names, and the values are the
      override names.
    permission_info: str, optional permission info that overrides default
      permission info group help.

  Returns:
    Presentation spec suitable for adding to concept parser.
  """

  if not permission_info:
    permission_info = '{} must hold permission {}'.format(
        "The 'Compute Engine Service Agent' service account",
        "'Cloud KMS CryptoKey Encrypter/Decrypter'")

  group_help = ('The Cloud KMS (Key Management Service) cryptokey that will be'
                ' used to protect the {}. {}.'.format(
                    resource, permission_info))

  presentation_spec = presentation_specs.ResourcePresentationSpec(
      '--kms-key',
      GetKmsKeyResourceSpec(region_fallthrough=region_fallthrough),
      group_help,
      flag_name_overrides=flag_overrides or {})
  return presentation_spec


def AddKmsKeyResourceArg(parser,
                         resource,
                         region_fallthrough=False,
                         flag_overrides=None,
                         permission_info=None,
                         required=False,
                         name='--kms-key'):
  """Add a resource argument for a KMS key to protect other resources.

  Args:
    parser: the parser for the command.
    resource: str, the name of the resource that the cryptokey will be used to
      protect.
    region_fallthrough: bool, True if the command has a region flag that should
      be used as a fallthrough for the kms location.
    flag_overrides: dict, The default flag names are 'kms-key', 'kms-keyring',
      'kms-location' and 'kms-project'. You can pass a dict of overrides where
      the keys of the dict are the default flag names, and the values are the
      override names.
    permission_info: str, optional permission info that overrides default
      permission info group help.
    required: bool, optional. True if the flag must be parsable by the parser.
      The default value is False.
    name: str, optional name of the arg for the KMS resource.
      Defaults to '--kms-key'.
  """
  if not permission_info:
    permission_info = '{} must hold permission {}'.format(
        "The 'Compute Engine Service Agent' service account",
        "'Cloud KMS CryptoKey Encrypter/Decrypter'")
  concept_parsers.ConceptParser.ForResource(
      name,
      GetKmsKeyResourceSpec(region_fallthrough=region_fallthrough),
      'The Cloud KMS (Key Management Service) cryptokey that will be used to '
      'protect the {}. {}.'.format(resource, permission_info),
      flag_name_overrides=flag_overrides,
      required=required).AddToParser(parser)


def AddKmsKeyResourceArgForKMS(parser, required, name):
  concept_parsers.ConceptParser.ForResource(
      name,
      GetKmsKeyResourceSpec(kms_prefix=False),
      'The KMS key resource.',
      required=required).AddToParser(parser)


def AddKmsKeyringResourceArgForKMS(parser, required, name):
  concept_parsers.ConceptParser.ForResource(
      name,
      GetKmsKeyRingResourceSpec(kms_prefix=False),
      'The KMS keyring resource.',
      required=required).AddToParser(parser)


def AddKmsEkmConnectionResourceArgForKMS(parser, required, name):
  concept_parsers.ConceptParser.ForResource(
      name,
      GetKmsEkmConnectionResourceSpec(kms_prefix=False),
      'The KMS ekm connection resource.',
      required=required).AddToParser(parser)


def AddKmsLocationResourceArgForKMS(parser, required, name):
  concept_parsers.ConceptParser.ForResource(
      name,
      GetKmsLocationResourceSpec(kms_prefix=False),
      'The KMS location resource.',
      required=required).AddToParser(parser)



# -*- coding: utf-8 -*- #
# Copyright 2014 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for accessing modules by installation independent paths."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import imp
import importlib
import os

from googlecloudsdk.core import exceptions
import six


class Error(exceptions.Error):
  """Exceptions for this module."""


class ImportModuleError(Error):
  """ImportModule failed."""


def ImportModule(module_path):
  """Imports a module object given its ModulePath and returns it.

  A module_path from GetModulePath() from any valid installation is importable
  by ImportModule() in another installation of same release.

  Args:
    module_path: The module path to import.

  Raises:
    ImportModuleError: Malformed module path or any failure to import.

  Returns:
    The Cloud SDK object named by module_path.
  """

  # First get the module.
  parts = module_path.split(':')
  if len(parts) > 2:
    raise ImportModuleError(
        'Module path [{}] must be in the form: '
        'package(.module)+(:attribute(.attribute)*)?'.format(module_path))
  try:
    module = importlib.import_module(parts[0])
  except ImportError as e:
    raise ImportModuleError(
        'Module path [{}] not found: {}.'.format(module_path, e))
  if len(parts) == 1:
    return module

  # March down the attributes to get the object within the module.
  obj = module
  attributes = parts[1].split('.')
  for attr in attributes:
    try:
      obj = getattr(obj, attr)
    except AttributeError as e:
      raise ImportModuleError(
          'Module path [{}] not found: {}.'.format(module_path, e))
  return obj


def _GetPrivateModulePath(module_path):
  """Mock hook that returns the module path for module that starts with '__'."""
  del module_path
  return None


def GetModulePath(obj):
  """Returns the module path string for obj, None if it's builtin.

  The module path is relative and importable by ImportModule() from any
  installation of the current release.

  Args:
    obj: The object to get the module path from.

  Returns:
    The module path name for obj if not builtin else None.
  """
  try:
    # An object either has a module ...
    module = obj.__module__
  except AttributeError:
    # ... or it has a __class__ that has a __module__.
    obj = obj.__class__
    module = obj.__module__
  if six.PY3 and module == 'builtins':
    return None
  # For example, '__builtins__' for Python 2 built-in types.
  if module.startswith('__'):
    module = _GetPrivateModulePath(module)  # pylint: disable=assignment-from-none, function is a test mock hook
    if not module:
      return None
  try:
    return module + ':' + obj.__name__
  except AttributeError:
    try:
      return module + ':' + obj.__class__.__name__
    except AttributeError:
      return None


def ImportPath(path):
  """Imports and returns the module given a python source file path."""
  module_dir = os.path.dirname(path)
  module_name = os.path.splitext(os.path.basename(path))[0]
  module_file = None
  try:
    module_file, module_path, module_description = imp.find_module(
        module_name, [module_dir])
    return imp.load_module(
        module_name, module_file, module_path, module_description)
  except ImportError as e:
    raise ImportModuleError(
        'Module file [{}] not found: {}.'.format(path, e))
  finally:
    if module_file:
      module_file.close()



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Troubleshoot Network and firewall for ssh connection."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import random
import socket
import string

import time

from dns import resolver

from googlecloudsdk.api_lib.services import enable_api
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.command_lib.compute import ssh_troubleshooter
from googlecloudsdk.core import log
from googlecloudsdk.core.console import console_io
from googlecloudsdk.core.console.console_io import OperationCancelledError

import six

_NUM_RANDOM_CHARACTERS = 5

_API_NETWORKMANAGEMENT_CLIENT_NAME = 'networkmanagement'
_API_COMPUTE_CLIENT_NAME = 'compute'
_API_CLIENT_VERSION_V1 = 'v1'

NETWORK_API = 'networkmanagement.googleapis.com'

networkmanagement_permissions = [
    'networkmanagement.connectivitytests.get',
]

CONNECTIVITY_TEST_MESSAGE = (
    'To view complete details of this test, see '
    'https://console.cloud.google.com/net-intelligence/connectivity/tests/details/{0}?project={1}\n'
    '\n'
    'Help for connectivity tests:\n'
    'https://cloud.google.com/network-intelligence-center/docs/connectivity-tests/concepts/overview\n')


def _GetRandomSuffix():
  random_characters = [
      random.choice(string.ascii_lowercase + string.digits)
      for _ in range(_NUM_RANDOM_CHARACTERS)]
  return ''.join(random_characters)


class NetworkTroubleshooter(ssh_troubleshooter.SshTroubleshooter):
  """Check network and firewall setting by running network connectivity test."""

  def __init__(self, project, zone, instance):
    self.project = project
    self.zone = zone
    self.instance = instance
    self.nm_client = apis.GetClientInstance(_API_NETWORKMANAGEMENT_CLIENT_NAME,
                                            _API_CLIENT_VERSION_V1)
    self.nm_message = apis.GetMessagesModule(_API_NETWORKMANAGEMENT_CLIENT_NAME,
                                             _API_CLIENT_VERSION_V1)
    self.compute_client = apis.GetClientInstance(_API_COMPUTE_CLIENT_NAME,
                                                 _API_CLIENT_VERSION_V1)
    self.compute_message = apis.GetMessagesModule(_API_COMPUTE_CLIENT_NAME,
                                                  _API_CLIENT_VERSION_V1)
    self.skip_troubleshoot = False
    self.test_id = 'ssh-troubleshoot-' + _GetRandomSuffix()

  def check_prerequisite(self):
    log.status.Print('---- Checking network connectivity ----')
    msg = ('The Network Management API is needed to check the VM\'s network '
           'connectivity.')
    prompt = 'Is it OK to enable it and check the VM\'s network connectivity?'
    cancel = ('Test skipped.\n'
              'To manually test network connectivity, try reaching another '
              'device on the same network.\n')
    try:
      prompt_continue = console_io.PromptContinue(
          message=msg,
          prompt_string=prompt,
          cancel_on_no=True,
          cancel_string=cancel)
      self.skip_troubleshoot = not prompt_continue
    except OperationCancelledError:
      self.skip_troubleshoot = True

    if self.skip_troubleshoot:
      return

    # Enable API
    enable_api.EnableService(self.project.name, NETWORK_API)
    # Test IAM Permission
    missing_permissions = self._CheckNetworkManagementPermissions()

    if missing_permissions:
      log.status.Print(
          'Missing the IAM permissions {0} necessary to perform the network '
          'connectivity test. To manually test network connectivity, try '
          'reaching another device on the same network.\n'.format(
              ' '.join(missing_permissions)))
      self.skip_troubleshoot = True
      return

  def cleanup_resources(self):
    return

  def troubleshoot(self):
    if self.skip_troubleshoot:
      return

    self.ip_address = self._GetSourceIPAddress()
    log.status.Print('Your source IP address is {0}\n'.format(self.ip_address))
    if not self.ip_address:
      log.status.Print('Could not resolve source external IP address, can\'t '
                       'run network connectivity test.\n')
      self.skip_troubleshoot = True
      return

    # Running network connectivity test if source IP available.
    operation_name = self._RunConnectivityTest()
    while not self._IsConnectivityTestFinish(operation_name):
      time.sleep(1)

    test_result = self._GetConnectivityTestResult()
    self._PrintConciseConnectivityTestResult(test_result)
    log.status.Print(CONNECTIVITY_TEST_MESSAGE.format(
        self.test_id, self.project.name))
    return

  def _RunConnectivityTest(self):
    connectivity_test = self._CreateConnectivityTest()
    # pylint: disable=line-too-long
    connectivity_test_create_req = self.nm_message.NetworkmanagementProjectsLocationsGlobalConnectivityTestsCreateRequest(
        parent='projects/{project_id}/locations/global'.format(
            project_id=self.project.name),
        testId=self.test_id,
        connectivityTest=connectivity_test)
    return self.nm_client.projects_locations_global_connectivityTests.Create(
        connectivity_test_create_req).name

  def _GetConnectivityTestResult(self):
    name = ('projects/{project_id}/locations/global/connectivityTests/'
            '{test_id}'.format(project_id=self.project.name,
                               test_id=self.test_id))
    connectivity_test_get_req = self.nm_message.NetworkmanagementProjectsLocationsGlobalConnectivityTestsGetRequest(
        name=name)
    return self.nm_client.projects_locations_global_connectivityTests.Get(
        connectivity_test_get_req)

  def _IsConnectivityTestFinish(self, name):
    # pylint: disable=line-too-long
    operation_get_req = self.nm_message.NetworkmanagementProjectsLocationsGlobalOperationsGetRequest(
        name=name)
    return self.nm_client.projects_locations_global_operations.Get(
        operation_get_req).done

  def _CreateConnectivityTest(self):
    return self.nm_message.ConnectivityTest(
        name='projects/{name}/locations/global/connectivityTests/{testId}'
        .format(name=self.project.name, testId=self.test_id),
        description='This connectivity test is created by '
        "'gcloud compute ssh --troubleshoot'",
        source=self.nm_message.Endpoint(
            ipAddress=self.ip_address, projectId=self.project.name),
        destination=self.nm_message.Endpoint(
            port=22,
            instance='projects/{project}/zones/{zone}/instances/{instance}'
            .format(
                project=self.project.name,
                zone=self.zone,
                instance=self.instance.name)),
        protocol='TCP')

  def _CheckNetworkManagementPermissions(self):
    resource_url = ('projects/{project_id}/locations/global/'
                    'connectivityTests/*'.format(project_id=self.project.name))
    test_permission_req = self.nm_message.TestIamPermissionsRequest(
        permissions=networkmanagement_permissions)
    # pylint: disable=line-too-long
    nm_testiampermission_req = self.nm_message.NetworkmanagementProjectsLocationsGlobalConnectivityTestsTestIamPermissionsRequest(
        resource=resource_url, testIamPermissionsRequest=test_permission_req)
    response = self.nm_client.projects_locations_global_connectivityTests.TestIamPermissions(nm_testiampermission_req)
    return set(networkmanagement_permissions) - set(response.permissions)

  def _GetSourceIPAddress(self):
    """Get current external IP from Google DNS server.

    Returns:
      str, an ipv4 address represented by string
    """
    re = resolver.Resolver()
    # pylint: disable=g-socket-gethostbyname
    re.nameservers = [socket.gethostbyname('ns1.google.com')]
    for rdata in re.query(qname='o-o.myaddr.l.google.com', rdtype='TXT'):
      return six.text_type(rdata).strip('"')
    return ''

  def _PrintConciseConnectivityTestResult(self, response):
    """Print concise network connectivity test result from response.

    Args:
      response: A response from projects_locations_global_connectivityTests Get

    Returns:

    """
    details = response.reachabilityDetails
    if details:
      log.status.Print('Network Connectivity Test Result: {0}\n'.format(
          details.result))



# -*- coding: utf-8 -*- #
# Copyright 2015 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cloud SDK markdown document renderer base class."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import abc
import io

from googlecloudsdk.core import log
from googlecloudsdk.core.resource import resource_printer

import six
from six.moves import range  # pylint: disable=redefined-builtin


# Font Attributes.
BOLD, ITALIC, CODE = list(range(3))


class TableColumnAttributes(object):
  """Markdown table column attributes.

  Attributes:
    align: Column alignment, one of {'left', 'center', 'right'}.
    label: Column heading label string.
    width: Minimum column width.
  """

  def __init__(self, align='left', label=None, width=0):
    self.align = align
    self.label = label
    self.width = width


class TableAttributes(object):
  """Markdown table attributes.

  Attributes:
    box: True if table and rows framed by box.
    columns: The list of column attributes.
    heading: The number of non-empty headings.
    margin: Extra margin to handle post-processing indent.
  """

  def __init__(self, box=False):
    self.box = box
    self.heading = 0
    self.columns = []

  def AddColumn(self, align='left', label='', width=0):
    """Adds the next column attributes to the table."""
    if label:
      self.heading += 1
    self.columns.append(
        TableColumnAttributes(align=align, label=label, width=width))

  def GetPrintFormat(self, margin=0, width=0):
    """Constructs and returns a resource_printer print format.

    Args:
      margin: Right hand side padding when one or more columns are wrapped.
      width: The table width.

    Returns:
      The resource printer format string.
    """
    fmt = ['table']
    attr = []
    if self.box:
      attr.append('box')
    if not self.heading:
      attr.append('no-heading')
    if margin:
      attr.append('margin={}'.format(margin))
    if width:
      attr.append('width={}'.format(width))
    if attr:
      fmt.append('[' + ','.join(attr) + ']')
    fmt.append('(')
    for index, column in enumerate(self.columns):
      if index:
        fmt.append(',')
      fmt.append('[{}]:label={}:align={}'.format(
          index, repr(column.label or '').lstrip('u'), column.align))
      if column.width:
        fmt.append(':width={}'.format(column.width))
    if margin:
      fmt.append(':wrap')
    fmt.append(')')
    return ''.join(fmt)


@six.add_metaclass(abc.ABCMeta)
class Renderer(object):
  r"""Markdown renderer base class.

  The member functions provide an abstract document model that matches markdown
  entities to output document renderings.

  Attributes:
    _blank: True if the output already contains a blank line. Used to avoid
      sequences of 2 or more blank lines in the output.
    _command: The command split into component names.
    _font: The font attribute bitmask.
    _indent: List of left indentations in characters indexed by _level.
    _lang: ```lang\n...\n``` code block language. None if not in code block,
      '' if in code block with no explicit lang specified.
    _level: The section or list level counting from 0.
    _out: The output stream.
    _title: The document title.
    _width: The output width in characters.
    command_metadata: Optional metadata of command.
    command_node: The command object that the document is being rendered for.
  """

  def __init__(self, out=None, title=None, width=80, command_metadata=None,
               command_node=None):
    self._blank = True
    self._command = ['gcloud']  # use command[0] instead of literal 'gcloud'
    self._font = 0
    self._indent = []
    self._lang = None
    self._level = 0
    self._out = out or log.out
    self._title = title
    self._width = width
    self.command_metadata = command_metadata
    self.command_node = command_node

  @property
  def command(self):
    """Returns the command split into component names."""
    return self._command

  def Blank(self):
    """The last output line is blank."""
    self._blank = True

  def Content(self):
    """Some non-blank line content was added to the output."""
    self._blank = False

  def HaveBlank(self):
    """Returns True if the last output line is blank."""
    return self._blank

  def Entities(self, buf):
    """Converts special characters to their entity tags.

    This is applied after font embellishments.

    Args:
      buf: The normal text that may contain special characters.

    Returns:
      The escaped string.
    """
    return buf

  def Escape(self, buf):
    """Escapes special characters in normal text.

    This is applied before font embellishments.

    Args:
      buf: The normal text that may contain special characters.

    Returns:
      The escaped string.
    """
    return buf

  def Finish(self):
    """Finishes all output document rendering."""
    return None

  def Font(self, unused_attr, unused_out=None):
    """Returns the font embellishment string for attr.

    Args:
      unused_attr: None to reset to the default font, otherwise one of BOLD,
        ITALIC, or CODE.
      unused_out: Writes tags line to this stream if not None.

    Returns:
      The font embellishment string.
    """
    return ''

  def SetCommand(self, command):
    """Sets the document command name.

    Args:
      command: The command split into component names.
    """
    self._command = command

  def SetLang(self, lang):
    """Sets the ```...``` code block language.

    Args:
      lang: The language name, None if not in a code block, '' is no explicit
        language specified.
    """
    self._lang = lang

  def Line(self):
    """Renders a paragraph separating line."""
    pass

  def Link(self, target, text):
    """Renders an anchor.

    Args:
      target: The link target URL.
      text: The text to be displayed instead of the link.

    Returns:
      The rendered link anchor and text.
    """
    if text:
      if target and '://' in target:
        # Show non-local targets.
        return '{0} ({1})'.format(text, target)
      return text
    if target:
      return target
    return '[]()'

  def LinkGlobalFlags(self, line):
    """Add global flags links to line if any.

    Args:
      line: The text line.

    Returns:
      line with annoted global flag links.
    """
    return line

  def TableLine(self, line, indent=0):
    """Adds an indented table line to the output.

    Args:
      line: The line to add. A newline will be added.
      indent: The number of characters to indent the table.
    """
    self._out.write(indent * ' ' + line + '\n')

  def Table(self, table, rows):
    """Renders a table.

    Nested tables are not supported.

    Args:
      table: A TableAttributes object.
      rows: A list of rows where each row is a list of column strings.
    """
    self.Line()
    indent = self._indent[self._level].indent + 2
    margin = indent if any([True for r in rows if ' ' in r[-1]]) else 0
    buf = io.StringIO()
    resource_printer.Print(
        rows,
        table.GetPrintFormat(margin=margin, width=self._width),
        out=buf)
    for line in buf.getvalue().split('\n')[:-1]:
      self.TableLine(line, indent=indent)
    self.Content()
    self.Line()



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A library that is used to support Functions commands."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import itertools

import enum

UNADVERTISED_PROVIDER_LABEL = 'unadvertised'


@enum.unique
class Resources(enum.Enum):

  class Resource(object):

    def __init__(self, name, collection_id):
      self.name = name
      self.collection_id = collection_id

  TOPIC = Resource('topic', 'pubsub.projects.topics')
  BUCKET = Resource('bucket', 'cloudfunctions.projects.buckets')
  FIREBASE_DB = Resource('firebase database', 'google.firebase.database.ref')
  FIRESTORE_DOC = Resource('firestore document', 'google.firestore.document')
  FIREBASE_ANALYTICS_EVENT = Resource('firebase analytics',
                                      'google.firebase.analytics.event')
  PROJECT = Resource('project', 'cloudresourcemanager.projects')


class TriggerProvider(object):
  """Represents --trigger-provider flag value options."""

  def __init__(self, label, events):
    self.label = label
    self.events = events
    for event in self.events:
      # Used to access provider properties when listing event types
      event.provider = self

  @property
  def default_event(self):
    return self.events[0]


class TriggerEvent(object):
  """Represents --trigger-event flag value options."""

  # Currently only project resource is optional
  OPTIONAL_RESOURCE_TYPES = [Resources.PROJECT]

  def __init__(self, label, resource_type):
    self.label = label
    self.resource_type = resource_type

  @property
  def event_is_optional(self):
    return self.provider.default_event == self

  # TODO(b/33097692) Let TriggerEvent know how to handle optional resources.
  @property
  def resource_is_optional(self):
    return self.resource_type in TriggerEvent.OPTIONAL_RESOURCE_TYPES


# TODO (b/73062780): Event types should not be hard-coded.
# Don't use those structures directly. Use registry object instead.
# By convention, first event type is default.
_PROVIDERS = [
    TriggerProvider('cloud.pubsub', [
        TriggerEvent('google.pubsub.topic.publish', Resources.TOPIC),
        TriggerEvent('providers/cloud.pubsub/eventTypes/topic.publish',
                     Resources.TOPIC),
    ]),
    TriggerProvider('cloud.storage', [
        TriggerEvent('google.storage.object.finalize', Resources.BUCKET),
        TriggerEvent('providers/cloud.storage/eventTypes/object.change',
                     Resources.BUCKET),
        TriggerEvent('google.storage.object.archive', Resources.BUCKET),
        TriggerEvent('google.storage.object.delete', Resources.BUCKET),
        TriggerEvent('google.storage.object.metadataUpdate', Resources.BUCKET),
    ]),
    TriggerProvider('google.firebase.database.ref', [
        TriggerEvent('providers/google.firebase.database/eventTypes/ref.create',
                     Resources.FIREBASE_DB),
        TriggerEvent('providers/google.firebase.database/eventTypes/ref.update',
                     Resources.FIREBASE_DB),
        TriggerEvent('providers/google.firebase.database/eventTypes/ref.delete',
                     Resources.FIREBASE_DB),
        TriggerEvent('providers/google.firebase.database/eventTypes/ref.write',
                     Resources.FIREBASE_DB),
    ]),
    TriggerProvider('google.firestore.document', [
        TriggerEvent('providers/cloud.firestore/eventTypes/document.create',
                     Resources.FIRESTORE_DOC),
        TriggerEvent('providers/cloud.firestore/eventTypes/document.update',
                     Resources.FIRESTORE_DOC),
        TriggerEvent('providers/cloud.firestore/eventTypes/document.delete',
                     Resources.FIRESTORE_DOC),
        TriggerEvent('providers/cloud.firestore/eventTypes/document.write',
                     Resources.FIRESTORE_DOC),
    ]),
    TriggerProvider('google.firebase.analytics.event', [
        TriggerEvent('providers/google.firebase.analytics/eventTypes/event.log',
                     Resources.FIREBASE_ANALYTICS_EVENT),
    ]),
    TriggerProvider('google.firebase.remoteConfig', [
        TriggerEvent('google.firebase.remoteconfig.update', Resources.PROJECT),
    ]),
    TriggerProvider('firebase.auth', [
        TriggerEvent('providers/firebase.auth/eventTypes/user.create',
                     Resources.PROJECT),
        TriggerEvent('providers/firebase.auth/eventTypes/user.delete',
                     Resources.PROJECT),
    ])
]


class _TriggerProviderRegistry(object):
  """This class encapsulates all Event Trigger related functionality."""

  def __init__(self, all_providers):
    self.providers = all_providers
    self._unadvertised_provider = TriggerProvider(UNADVERTISED_PROVIDER_LABEL,
                                                  [])

  def ProvidersLabels(self):
    return (p.label for p in self.providers)

  def Provider(self, provider):
    return next((p for p in self.providers if p.label == provider))

  def EventsLabels(self, provider):
    return (e.label for e in self.Provider(provider).events)

  def AllEventLabels(self):
    all_events = (self.EventsLabels(p.label) for p in self.providers)
    return itertools.chain.from_iterable(all_events)

  def Event(self, provider, event):
    return next((e for e in self.Provider(provider).events if e.label == event))

  def ProviderForEvent(self, event_label):
    for p in self.providers:
      if event_label in self.EventsLabels(p.label):
        return p
    return self._unadvertised_provider


TRIGGER_PROVIDER_REGISTRY = _TriggerProviderRegistry(_PROVIDERS)



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Holds iterator for reading from stdin."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import sys

import six


class StdinIterator(six.Iterator):
  """An iterator that returns lines from stdin."""

  def __iter__(self):
    return self

  def __next__(self):
    line = sys.stdin.readline()
    if not line:
      raise StopIteration
    return line.rstrip()


def get_urls_iterable(normal_urls_argument, should_read_paths_from_stdin):
  """Helps command decide between normal URL args and a StdinIterator."""
  if not (normal_urls_argument or should_read_paths_from_stdin):
    raise ValueError('Must have URL arguments if not reading paths from stdin.')
  if normal_urls_argument and should_read_paths_from_stdin:
    raise ValueError(
        'Cannot have both read from stdin flag and normal URL arguments.')
  if should_read_paths_from_stdin:
    return StdinIterator()
  return normal_urls_argument



# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for Cloud Firestore commands."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from apitools.base.py import encoding
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.core.util import text


FIRESTORE_INDEX_API_VERSION = 'v1'


def GetMessagesModule():
  return apis.GetMessagesModule('firestore', FIRESTORE_INDEX_API_VERSION)


def GetDatabaseFallthrough():
  """Python hook to get the value for the default database.

  Firestore currently only supports one database called '(default)'.

  Returns:
    The name of the default database.
  """
  return '(default)'


def GetCollectionGroupFallthrough():
  """Python hook to get the value for the '-' collection group.

  See details at:

  https://cloud.google.com/apis/design/design_patterns#list_sub-collections

  This allows us to describe or delete an index by specifying just its ID,
  without needing to know which collection group it belongs to.

  Returns:
    The value of the wildcard collection group.
  """
  return '-'


def GetDefaultFieldCollectionGroupFallthrough():
  return '__default__'


def GetDefaultFieldPathFallthrough():
  return '*'


def ValidateFieldConfig(unused_ref, args, request):
  """Python hook to validate the field configuration of the given request.

  Note that this hook is only called after the request has been formed based on
  the spec. Thus, the validation of the user's choices for order and
  array-config, as well as the check for the required field-path attribute, have
  already been performed. As such the only remaining things to verify are that
  the user has specified at least 2 fields, and that exactly one of order or
  array-config was specified for each field.

  Args:
    unused_ref: The resource ref (unused).
    args: The parsed arg namespace.
    request: The request formed based on the spec.
  Returns:
    The original request assuming the field configuration is valid.
  Raises:
    InvalidArgumentException: If the field configuration is invalid.
  """
  if len(args.field_config) < 2:
    raise exceptions.InvalidArgumentException(
        '--field-config',
        'Composite indexes must be configured with at least 2 fields. For '
        'single-field index management, use the commands under `gcloud '
        'firestore indexes fields`.')

  invalid_field_configs = []
  for field_config in args.field_config:
    # Because of the way declarative ArgDict parsing works, the type of
    # field_config here is already an apitools message, as opposed to an
    # ArgDict.
    order = field_config.order
    array_config = field_config.arrayConfig
    if (order and array_config) or (not order and not array_config):
      invalid_field_configs.append(field_config)

  if invalid_field_configs:
    raise exceptions.InvalidArgumentException(
        '--field-config',
        "Exactly one of 'order' or 'array-config' must be specified for the "
        "{field_word} with the following {path_word}: [{paths}].".format(
            field_word=text.Pluralize(len(invalid_field_configs), 'field'),
            path_word=text.Pluralize(len(invalid_field_configs), 'path'),
            paths=', '.join(field_config.fieldPath
                            for field_config in invalid_field_configs)))

  return request


def ExtractOperationMetadata(response, unused_args):
  """Python hook to extract the operation metadata message.

  This is needed because apitools gives us a MetadataValue message for the
  operation metadata field, instead of the actual message that we want.

  Args:
    response: The response field in the operation returned by the API.
    unused_args: The parsed arg namespace (unused).
  Returns:
    The metadata field converted to a
    GoogleFirestoreAdminV1IndexOperationMetadata message.
  """
  messages = GetMessagesModule()
  return encoding.DictToMessage(
      encoding.MessageToDict(response.metadata),
      messages.GoogleFirestoreAdminV1IndexOperationMetadata)


def ValidateFieldArg(ref, unused_args, request):
  """Python hook to validate that the field reference is correctly specified.

  The user should be able to describe database-wide settings as well as
  collection-group wide settings; however it doesn't make sense to describe a
  particular field path's settings unless the collection group was also
  specified. The API will catch this but it's better to do it here for a clearer
  error message.

  Args:
    ref: The field resource reference.
    unused_args: The parsed arg namespace (unused).
    request: The field describe request.
  Returns:
    The original request assuming the field configuration is valid.
  Raises:
    InvalidArgumentException: If the field resource is invalid.
  """
  if (ref.fieldsId != GetDefaultFieldPathFallthrough() and
      ref.collectionGroupsId == GetDefaultFieldCollectionGroupFallthrough()):
    raise exceptions.InvalidArgumentException(
        'FIELD',
        'Collection group must be provided if the field path was specified.')
  return request


def CreateIndexMessage(messages, index):
  """Creates a message for the given index.

  Args:
    messages: The Cloud Firestore messages module.
    index: The index ArgDict.
  Returns:
    GoogleFirestoreAdminV1Index
  """
  # Currently all indexes are COLLECTION-scoped
  query_scope = (
      messages.GoogleFirestoreAdminV1Index.QueryScopeValueValuesEnum.COLLECTION)
  # Since this is a single-field index there will only be 1 field
  index_fields = [
      messages.GoogleFirestoreAdminV1IndexField(
          arrayConfig=index.get('array-config'), order=index.get('order'))
  ]

  return messages.GoogleFirestoreAdminV1Index(
      queryScope=query_scope, fields=index_fields)


def ValidateFieldIndexArgs(args):
  """Validates the repeated --index arg.

  Args:
    args: The parsed arg namespace.
  Raises:
    InvalidArgumentException: If the provided indexes are incorrectly specified.
  """
  if not args.IsSpecified('index'):
    return

  for index in args.index:
    order = index.get('order')
    array_config = index.get('array-config')
    if (order and array_config) or (not order and not array_config):
      raise exceptions.InvalidArgumentException(
          '--index',
          "Exactly one of 'order' or 'array-config' must be specified "
          "for each --index flag provided.")


def UpdateFieldRequestTtls(ref, args, request):
  """Update field request for TTL.

  Args:
    ref: The field resource reference(unused).
    args: The parsed arg namespace.
    request: The ttl field request.
  Raises:
    InvalidArgumentException: If the provided indexes are incorrectly specified.
  Returns:
    UpdateFieldRequest
  """
  messages = GetMessagesModule()
  request.updateMask = 'ttlConfig'
  ttl_config = None

  if args.enable_ttl:
    ttl_config = messages.GoogleFirestoreAdminV1TtlConfig()

  request.googleFirestoreAdminV1Field = messages.GoogleFirestoreAdminV1Field(
      name=ref.RelativeName(),
      ttlConfig=ttl_config)

  return request


def CreateFieldUpdateRequest(ref, args):
  """Python hook to create the field update request.

  The mapping of index config message to API behavior is as follows:
    None          - Clears the exemption
    indexes=[]    - Disables all indexes
    indexes=[...] - Sets the index config to the indexes provided

  Args:
    ref: The field resource reference.
    args: The parsed arg namespace.
  Returns:
    FirestoreProjectsDatabasesCollectionGroupsFieldsPatchRequest
  """
  ValidateFieldIndexArgs(args)

  messages = GetMessagesModule()
  index_config = None
  if args.disable_indexes:
    index_config = messages.GoogleFirestoreAdminV1IndexConfig(indexes=[])
  elif args.IsSpecified('index'):
    # args.index is a repeated argument
    index_config = messages.GoogleFirestoreAdminV1IndexConfig(
        indexes=[CreateIndexMessage(messages, index) for index in args.index])

  field = messages.GoogleFirestoreAdminV1Field(
      name=ref.RelativeName(), indexConfig=index_config)

  request = (
      messages.FirestoreProjectsDatabasesCollectionGroupsFieldsPatchRequest(
          name=ref.RelativeName(),
          updateMask='indexConfig',
          googleFirestoreAdminV1Field=field))

  return request


def GetDatabaseType(value):
  """Return API required format for type specified by value."""
  messages = GetMessagesModule()
  if value == 'firestore-native':
    return messages.GoogleFirestoreAdminV1Database.TypeValueValuesEnum.FIRESTORE_NATIVE
  elif value == 'datastore-mode':
    return messages.GoogleFirestoreAdminV1Database.TypeValueValuesEnum.DATASTORE_MODE
  else:
    raise exceptions.InvalidArgumentException('invalid type: ' + value)


def CreateDatabaseUpdateRequest(ref, args):
  """Create a database update request.

  Args:
    ref: The resource ref.
    args: The parsed arg namespace.

  Returns:
    A new database update request with new retention.
  """
  messages = GetMessagesModule()
  database = messages.GoogleFirestoreAdminV1Database(
      name=ref.RelativeName(),
      type=GetDatabaseType(args.type))
  request = (
      messages.FirestoreProjectsDatabasesPatchRequest(
          name=ref.RelativeName(),
          updateMask='type',
          googleFirestoreAdminV1Database=database))
  return request



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for AI Platform endpoints commands."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import io
import json

from googlecloudsdk.command_lib.ai import errors
from googlecloudsdk.core import resources
from googlecloudsdk.core import yaml
from googlecloudsdk.core.console import console_io
from googlecloudsdk.core.util import encoding

import six


def ParseOperation(operation_name):
  """Parse operation resource to the operation reference object.

  Args:
    operation_name: The operation resource to wait on

  Returns:
    The operation reference object
  """
  if '/endpoints/' in operation_name:
    try:
      return resources.REGISTRY.ParseRelativeName(
          operation_name,
          collection='aiplatform.projects.locations.endpoints.operations')
    except resources.WrongResourceCollectionException:
      pass
  return resources.REGISTRY.ParseRelativeName(
      operation_name, collection='aiplatform.projects.locations.operations')


def ReadRequest(input_file):
  """Reads a JSON request from the specified input file.

  Args:
    input_file: An open file-like object for the input file.

  Returns:
    A json object from the input file

  Raises:
    InvalidInstancesFileError: If the input file is invalid.
  """
  try:
    request = yaml.load(input_file)
  except ValueError:
    raise errors.InvalidInstancesFileError(
        'Input instances are not in JSON format. '
        'See `gcloud ai endpoints predict --help` for details.')

  if not isinstance(request, dict):
    raise errors.InvalidInstancesFileError(
        'Input instances are not in JSON format. '
        'See `gcloud ai endpoints predict --help` for details.')

  if 'instances' not in request:
    raise errors.InvalidInstancesFileError(
        'Invalid JSON request: missing "instances" attribute')

  if not isinstance(request['instances'], list):
    raise errors.InvalidInstancesFileError(
        'Invalid JSON request: "instances" must be a list')

  return request


def ReadInstancesFromArgs(json_request):
  """Reads the instances from the given file path ('-' for stdin).

  Args:
    json_request: str or None, a path to a file ('-' for stdin) containing
        the JSON body of a prediction request.

  Returns:
    A list of instances.

  Raises:
    InvalidInstancesFileError: If the input file is invalid (invalid format or
        contains too many/zero instances), or an improper combination of input
        files was given.
  """
  data = console_io.ReadFromFileOrStdin(json_request, binary=True)
  with io.BytesIO(data) as f:
    return ReadRequest(f)


def GetDefaultFormat(predictions, key_name='predictions'):
  """Get default output format for prediction results."""
  if not isinstance(predictions, list):
    # This usually indicates some kind of error case, so surface the full API
    # response
    return 'json'
  elif not predictions:
    return None
  # predictions is guaranteed by API contract to be a list of similarly shaped
  # objects, but we don't know ahead of time what those objects look like.
  elif isinstance(predictions[0], dict):
    keys = ', '.join(sorted(predictions[0].keys()))
    return """
          table(
              {}:format="table(
                  {}
              )"
          )""".format(key_name, keys)

  else:
    return 'table[no-heading]({})'.format(key_name)
